/* (C) 2009-2011 Yana */

#include "yk0a_internal.h"
#include "yk0a.h"
#include "yk0a_base.h"
#include "yk0a_interrupt.h"
#include "yk0a_setup.h"
#include <stddef.h>

/* queue */

typedef struct yk0a_queue {
	struct yk0a_queue *next;
	struct yk0a_queue *previous;
} yk0a_queue_t;

static void yk0a_queue_append(yk0a_queue_t *root, yk0a_queue_t *node) {
	node->next = root;
	node->previous = root->previous;
	root->previous->next = node;
	root->previous = node;
}

static yk0a_queue_t *yk0a_queue_copy(yk0a_queue_t *root) {
	if (root->next == root) {
		return NULL;
	} else {
		return root->next;
	}
}

static void yk0a_queue_delete(yk0a_queue_t *node) {
	node->next->previous = node->previous;
	node->previous->next = node->next;
}

static void yk0a_queue_initialize(yk0a_queue_t *root) {
	root->next = root;
	root->previous = root;
}

static yk0a_queue_t *yk0a_queue_remove(yk0a_queue_t *root) {
	if (root->next == root) {
		return NULL;
	} else {
		yk0a_queue_t *node;
		node = root->next;
		root->next = node->next;
		node->next->previous = node->previous;
		return node;
	}
}

/* type */

typedef yk0a_uint32_t yk0a_state_t;

typedef struct yk0a_task_control {
	yk0a_queue_t queue;
	yk0a_uint32_t context[YK0A_CONTEXT_SIZE];
	yk0a_priority_t priority;
	yk0a_state_t state;
	union {
		yk0a_port_t port;
		struct yk0a_task_control *replyer;
	} u;
} yk0a_task_control_t;

typedef yk0a_task_control_t *yk0a_task_link_t;

typedef struct {
	yk0a_queue_t queue;
	yk0a_flag_t flag;
	yk0a_task_link_t receiver;
} yk0a_port_control_t;

/* variable */

yk0a_port_control_t yk0a_port_control[YK0A_SETUP_PORT_SIZE];
yk0a_task_link_t yk0a_schedule_link;
yk0a_queue_t yk0a_schedule_queue[YK0A_SETUP_PRIORITY_SIZE];
yk0a_task_control_t yk0a_task_control[YK0A_SETUP_TASK_SIZE];

/* access function */

static void yk0a_port_append(yk0a_port_t port, yk0a_task_link_t sender) {
	yk0a_queue_append((yk0a_queue_t *)&yk0a_port_control[port], (yk0a_queue_t *)sender);
	yk0a_port_control[port].flag |= YK0A_FLAG_KERNEL_SENDER;
}

static void yk0a_port_delete(yk0a_port_t port, yk0a_task_link_t sender) {
	yk0a_queue_delete((yk0a_queue_t *)sender);
	if (yk0a_queue_copy((yk0a_queue_t *)&yk0a_port_control[port]) == NULL) {
		yk0a_port_control[port].flag &= ~YK0A_FLAG_KERNEL_SENDER;
	}
}

static void yk0a_port_initialize(void) {
	yk0a_port_t port;
	for (port = 0; port < YK0A_SETUP_PORT_SIZE; ++port) {
		yk0a_queue_initialize((yk0a_queue_t *)&yk0a_port_control[port]);
		yk0a_port_control[port].flag = 0;
	}
}

static yk0a_flag_t yk0a_port_read_flag(yk0a_port_t port) {
	return yk0a_port_control[port].flag;
}

static yk0a_task_link_t yk0a_port_read_receiver(yk0a_port_t port) {
	return yk0a_port_control[port].receiver;
}

static yk0a_task_link_t yk0a_port_remove(yk0a_port_t port) {
	yk0a_task_link_t sender;
	sender = (yk0a_task_link_t)yk0a_queue_remove((yk0a_queue_t *)&yk0a_port_control[port]);
	if (yk0a_queue_copy((yk0a_queue_t *)&yk0a_port_control[port]) == NULL) {
		yk0a_port_control[port].flag &= ~YK0A_FLAG_KERNEL_SENDER;
	}
	return sender;
}

static void yk0a_port_write_flag(yk0a_port_t port, yk0a_flag_t flag) {
	yk0a_port_control[port].flag = flag;
}

static void yk0a_port_write_receiver(yk0a_port_t port, yk0a_task_link_t receiver) {
	yk0a_port_control[port].receiver = receiver;
}

static void yk0a_schedule_append(yk0a_task_link_t link) {
	yk0a_queue_append(&yk0a_schedule_queue[link->priority], (yk0a_queue_t *)link);
}

static void yk0a_schedule_delete(yk0a_task_link_t link) {
	yk0a_queue_delete((yk0a_queue_t *)link);
}

static void yk0a_schedule_initialize(void) {
	yk0a_priority_t priority;
	yk0a_schedule_link = NULL;
	for (priority = 0; priority < YK0A_SETUP_PRIORITY_SIZE; ++priority) {
		yk0a_queue_initialize(&yk0a_schedule_queue[priority]);
	}
}

static void yk0a_schedule_update(void) {
	yk0a_priority_t priority;
	for (priority = 0; priority < YK0A_SETUP_PRIORITY_SIZE; ++priority) {
		yk0a_schedule_link = (yk0a_task_link_t)yk0a_queue_copy(&yk0a_schedule_queue[priority]);
		if (yk0a_schedule_link != NULL) {
			return;
		}
	}
}

static yk0a_task_t yk0a_task_link_to_task(yk0a_task_link_t link) {
	return link - &yk0a_task_control[0];
}

static yk0a_port_t yk0a_task_read_port(yk0a_task_link_t link) {
	return link->u.port;
}

static yk0a_priority_t yk0a_task_read_priority(yk0a_task_link_t link) {
	return link->priority;
}

static yk0a_task_link_t yk0a_task_read_replyer(yk0a_task_link_t link) {
	return link->u.replyer;
}

static yk0a_state_t yk0a_task_read_state(yk0a_task_link_t link) {
	return link->state;
}

static yk0a_task_link_t yk0a_task_to_task_link(yk0a_task_t task) {
	return &yk0a_task_control[task];
}

static void yk0a_task_write_port(yk0a_task_link_t link, yk0a_port_t port) {
	link->u.port = port;
}

static void yk0a_task_write_priority(yk0a_task_link_t link, yk0a_priority_t priority) {
	link->priority = priority;
}

static void yk0a_task_write_replyer(yk0a_task_link_t link, yk0a_task_link_t replyer) {
	link->u.replyer = replyer;
}

static void yk0a_task_write_state(yk0a_task_link_t link, yk0a_state_t state) {
	link->state = state;
}

/* context function */

static yk0a_context_value_t yk0a_task_read_context(yk0a_task_link_t link, yk0a_context_type_t type) {
	return link->context[type];
}

static void yk0a_task_write_context(yk0a_task_link_t link, yk0a_context_type_t type, yk0a_context_value_t value) {
	link->context[type] = value;
}

static yk0a_port_t yk0a_accept_read_port(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static void yk0a_accept_write_task(yk0a_task_link_t link, yk0a_task_t task) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = task;
}

static yk0a_port_t yk0a_bit_flag_read_port(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_flag_t yk0a_bit_flag_read_flag_off(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static yk0a_flag_t yk0a_bit_flag_read_flag_on(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT3];
}

static void yk0a_bit_flag_write_flag(yk0a_task_link_t link, yk0a_flag_t flag) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = flag;
}

static yk0a_port_t yk0a_call_read_port(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_port_t yk0a_connect_read_task_replyee(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_task_t yk0a_connect_read_task_replyer(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static yk0a_port_t yk0a_forward_read_port(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_task_t yk0a_forward_read_task(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static yk0a_task_t yk0a_get_context_read_task(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_context_type_t yk0a_get_context_read_type(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static void yk0a_get_context_write_value(yk0a_task_link_t link, yk0a_context_value_t value) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = value;
}

static yk0a_task_t yk0a_get_priority_read_task(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static void yk0a_get_priority_write_priority(yk0a_task_link_t link, yk0a_priority_t priority) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = priority;
}

static yk0a_port_t yk0a_listen_read_port(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_flag_t yk0a_listen_read_flag(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static void yk0a_listen_write_flag(yk0a_task_link_t link, yk0a_flag_t flag) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = flag;
}

static yk0a_task_t yk0a_reply_read_task(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_task_t yk0a_set_context_read_task(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_context_type_t yk0a_set_context_read_type(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static yk0a_context_value_t yk0a_set_context_read_value(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT3];
}

static void yk0a_set_context_write_value(yk0a_task_link_t link, yk0a_context_value_t value) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = value;
}

static yk0a_task_t yk0a_set_priority_read_task(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_priority_t yk0a_set_priority_read_priority(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static void yk0a_set_priority_write_priority(yk0a_task_link_t link, yk0a_priority_t priority) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = priority;
}

static yk0a_interrupt_t yk0a_interrupt_bit_mask_read_mask_off(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_interrupt_t yk0a_interrupt_bit_mask_read_mask_on(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static void yk0a_interrupt_bit_mask_write_mask(yk0a_task_link_t link, yk0a_interrupt_t mask) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = mask;
}

static yk0a_interrupt_t yk0a_interrupt_bit_request_read_request_off(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT1];
}

static yk0a_interrupt_t yk0a_interrupt_bit_request_read_request_on(yk0a_task_link_t link) {
	return link->context[YK0A_CONTEXT_ARGUMENT2];
}

static void yk0a_interrupt_bit_request_write_request(yk0a_task_link_t link, yk0a_interrupt_t request) {
	link->context[YK0A_CONTEXT_ARGUMENT1] = request;
}

/* service function */

static void yk0a_exception_forward(yk0a_port_t port, yk0a_task_link_t link) {
	yk0a_schedule_delete(link);
	yk0a_task_write_port(link, port);
	yk0a_task_write_state(link, YK0A_STATE_RECEIVE);
	yk0a_port_append(port, link);
	if ((yk0a_port_read_flag(port) & YK0A_FLAG_KERNEL_RECEIVER) != 0 && (yk0a_listen_read_flag(yk0a_port_read_receiver(port)) & YK0A_FLAG_KERNEL_SENDER) != 0) {
		yk0a_flag_t f;
		f = yk0a_listen_read_flag(yk0a_port_read_receiver(port));
		yk0a_listen_write_flag(yk0a_port_read_receiver(port), yk0a_port_read_flag(port));
		yk0a_port_write_flag(port, yk0a_port_read_flag(port) & (~(f | YK0A_FLAG_KERNEL_RECEIVER) | YK0A_FLAG_KERNEL_SENDER));
		yk0a_task_write_state(yk0a_port_read_receiver(port), YK0A_STATE_READY);
		yk0a_schedule_append(yk0a_port_read_receiver(port));
	}
	yk0a_schedule_update();
}

static void yk0a_interrupt_bit_flag(yk0a_port_t port, yk0a_flag_t flag_off, yk0a_flag_t flag_on) {
	if ((yk0a_port_read_flag(port) & YK0A_FLAG_KERNEL_RECEIVER) == 0 || (yk0a_listen_read_flag(yk0a_port_read_receiver(port)) & flag_on) == 0) {
		yk0a_port_write_flag(port, (yk0a_port_read_flag(port) & ~flag_off) | flag_on);
	} else {
		yk0a_flag_t f;
		f = yk0a_listen_read_flag(yk0a_port_read_receiver(port));
		yk0a_listen_write_flag(yk0a_port_read_receiver(port), (yk0a_port_read_flag(port) & ~flag_off) | flag_on);
		yk0a_port_write_flag(port, ((yk0a_port_read_flag(port) & ~flag_off) | flag_on) & (~(f | YK0A_FLAG_KERNEL_RECEIVER) | YK0A_FLAG_KERNEL_SENDER));
		yk0a_task_write_state(yk0a_port_read_receiver(port), YK0A_STATE_READY);
		yk0a_schedule_append(yk0a_port_read_receiver(port));
		yk0a_schedule_update();
	}
}

static yk0a_error_t yk0a_procedure_accept(yk0a_task_link_t link) {
	yk0a_task_link_t l;
	yk0a_port_t p;
	p = yk0a_accept_read_port(link);
	if (p >= YK0A_SETUP_PORT_SIZE) {
		return 1;
	}
	if ((yk0a_port_read_flag(p) & YK0A_FLAG_KERNEL_SENDER) == 0) {
		return 1;
	}
	l = yk0a_port_remove(p);
	yk0a_task_write_replyer(l, link);
	yk0a_task_write_state(l, YK0A_STATE_REPLY);
	yk0a_accept_write_task(link, yk0a_task_link_to_task(l));
	return 0;
}

static yk0a_error_t yk0a_procedure_bit_flag(yk0a_task_link_t link) {
	yk0a_flag_t off;
	yk0a_flag_t on;
	yk0a_port_t p;
	p = yk0a_bit_flag_read_port(link);
	if (p >= YK0A_SETUP_PORT_SIZE) {
		return 1;
	}
	off = yk0a_bit_flag_read_flag_off(link);
	on = yk0a_bit_flag_read_flag_on(link);
	yk0a_bit_flag_write_flag(link, yk0a_port_read_flag(p));
	yk0a_interrupt_bit_flag(p, off & ~YK0A_FLAG_KERNEL_MASK, on & ~YK0A_FLAG_KERNEL_MASK);
	return 0;
}

static yk0a_error_t yk0a_procedure_call(yk0a_task_link_t link) {
	yk0a_port_t p;
	p = yk0a_call_read_port(link);
	if (p >= YK0A_SETUP_PORT_SIZE) {
		return 1;
	}
	yk0a_exception_forward(p, link);
	return 0;
}

static yk0a_error_t yk0a_procedure_connect(yk0a_task_link_t link) {
	yk0a_task_t replyee;
	yk0a_task_t replyer;
	replyee = yk0a_connect_read_task_replyee(link);
	if (replyee >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0a_task_read_state(yk0a_task_to_task_link(replyee)) != YK0A_STATE_REPLY) {
		return 1;
	}
	if (yk0a_task_read_replyer(yk0a_task_to_task_link(replyee)) != link) {
		return 1;
	}
	replyer = yk0a_connect_read_task_replyer(link);
	if (replyer >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0a_task_read_state(yk0a_task_to_task_link(replyer)) != YK0A_STATE_REPLY) {
		return 1;
	}
	if (yk0a_task_read_replyer(yk0a_task_to_task_link(replyer)) != link) {
		return 1;
	}
	yk0a_task_write_replyer(yk0a_task_to_task_link(replyee), yk0a_task_to_task_link(replyer));
	return 0;
}

static yk0a_error_t yk0a_procedure_forward(yk0a_task_link_t link) {
	yk0a_port_t p;
	yk0a_task_t t;
	p = yk0a_forward_read_port(link);
	if (p >= YK0A_SETUP_PORT_SIZE) {
		return 1;
	}
	t = yk0a_forward_read_task(link);
	if (t >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	switch (yk0a_task_read_state(yk0a_task_to_task_link(t))) {
	case YK0A_STATE_READY:
		if (yk0a_task_to_task_link(t) == link) {
			return 1;
		}
		yk0a_schedule_delete(yk0a_task_to_task_link(t));
		break;
	case YK0A_STATE_RECEIVE:
		yk0a_port_delete(yk0a_task_read_port(yk0a_task_to_task_link(t)), yk0a_task_to_task_link(t));
		break;
	case YK0A_STATE_REPLY:
		if (yk0a_task_read_replyer(yk0a_task_to_task_link(t)) != link) {
			return 1;
		}
		break;
	case YK0A_STATE_SEND:
		yk0a_listen_write_flag(yk0a_task_to_task_link(t), 0);
		yk0a_port_write_flag(yk0a_task_read_port(yk0a_task_to_task_link(t)), yk0a_port_read_flag(yk0a_task_read_port(yk0a_task_to_task_link(t))) & ~YK0A_FLAG_KERNEL_RECEIVER);
		break;
	}
	yk0a_task_write_port(yk0a_task_to_task_link(t), p);
	yk0a_task_write_state(yk0a_task_to_task_link(t), YK0A_STATE_RECEIVE);
	yk0a_port_append(p, yk0a_task_to_task_link(t));
	if ((yk0a_port_read_flag(p) & YK0A_FLAG_KERNEL_RECEIVER) != 0 && (yk0a_listen_read_flag(yk0a_port_read_receiver(p)) & YK0A_FLAG_KERNEL_SENDER) != 0) {
		yk0a_flag_t f;
		f = yk0a_listen_read_flag(yk0a_port_read_receiver(p));
		yk0a_listen_write_flag(yk0a_port_read_receiver(p), yk0a_port_read_flag(p));
		yk0a_port_write_flag(p, yk0a_port_read_flag(p) & (~(f | YK0A_FLAG_KERNEL_RECEIVER) | YK0A_FLAG_KERNEL_SENDER));
		yk0a_task_write_state(yk0a_port_read_receiver(p), YK0A_STATE_READY);
		yk0a_schedule_append(yk0a_port_read_receiver(p));
		yk0a_schedule_update();
	}
	return 0;
}

static yk0a_error_t yk0a_procedure_get_context(yk0a_task_link_t link) {
	yk0a_context_type_t ct;
	yk0a_task_t t;
	t = yk0a_get_context_read_task(link);
	if (t >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0a_task_read_state(yk0a_task_to_task_link(t)) != YK0A_STATE_REPLY) {
		return 1;
	}
	if (yk0a_task_read_replyer(yk0a_task_to_task_link(t)) != link) {
		return 1;
	}
	ct = yk0a_get_context_read_type(link);
	if (ct >= YK0A_CONTEXT_SIZE) {
		return 1;
	}
	yk0a_get_context_write_value(link, yk0a_task_read_context(yk0a_task_to_task_link(t), ct));
	return 0;
}

static yk0a_error_t yk0a_procedure_get_priority(yk0a_task_link_t link) {
	yk0a_task_t t;
	t = yk0a_get_priority_read_task(link);
	if (t >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	yk0a_get_priority_write_priority(link, yk0a_task_read_priority(yk0a_task_to_task_link(t)));
	return 0;
}

static yk0a_error_t yk0a_procedure_listen(yk0a_task_link_t link) {
	yk0a_flag_t f;
	yk0a_port_t p;
	p = yk0a_listen_read_port(link);
	if (p >= YK0A_SETUP_PORT_SIZE) {
		return 1;
	}
	if ((yk0a_port_read_flag(p) & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
		return 1;
	}
	f = yk0a_listen_read_flag(link);
	if (((yk0a_port_read_flag(p) | YK0A_FLAG_KERNEL_RECEIVER) & f) != 0) {
		yk0a_listen_write_flag(link, yk0a_port_read_flag(p) | YK0A_FLAG_KERNEL_RECEIVER);
		yk0a_port_write_flag(p, yk0a_port_read_flag(p) & (~f | YK0A_FLAG_KERNEL_SENDER));
	} else {
		yk0a_schedule_delete(link);
		yk0a_task_write_port(link, p);
		yk0a_task_write_state(link, YK0A_STATE_SEND);
		yk0a_port_write_flag(p, yk0a_port_read_flag(p) | YK0A_FLAG_KERNEL_RECEIVER);
		yk0a_port_write_receiver(p, link);
		yk0a_schedule_update();
	}
	return 0;
}

static yk0a_error_t yk0a_procedure_reply(yk0a_task_link_t link) {
	yk0a_task_t t;
	t = yk0a_reply_read_task(link);
	if (t >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0a_task_read_state(yk0a_task_to_task_link(t)) != YK0A_STATE_REPLY) {
		return 1;
	}
	if (yk0a_task_read_replyer(yk0a_task_to_task_link(t)) != link) {
		return 1;
	}
	yk0a_task_write_state(yk0a_task_to_task_link(t), YK0A_STATE_READY);
	yk0a_schedule_append(yk0a_task_to_task_link(t));
	yk0a_schedule_update();
	return 0;
}

static yk0a_error_t yk0a_procedure_set_context(yk0a_task_link_t link) {
	yk0a_context_type_t ct;
	yk0a_context_value_t cv;
	yk0a_task_t t;
	t = yk0a_set_context_read_task(link);
	if (t >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0a_task_read_state(yk0a_task_to_task_link(t)) != YK0A_STATE_REPLY) {
		return 1;
	}
	if (yk0a_task_read_replyer(yk0a_task_to_task_link(t)) != link) {
		return 1;
	}
	ct = yk0a_set_context_read_type(link);
	if (ct >= YK0A_CONTEXT_SIZE) {
		return 1;
	}
	cv = yk0a_set_context_read_value(link);
	yk0a_set_context_write_value(link, yk0a_task_read_context(yk0a_task_to_task_link(t), ct));
	yk0a_task_write_context(yk0a_task_to_task_link(t), ct, cv);
	return 0;
}

static yk0a_error_t yk0a_procedure_set_priority(yk0a_task_link_t link) {
	yk0a_priority_t p;
	yk0a_task_t t;
	t = yk0a_set_priority_read_task(link);
	if (t >= YK0A_SETUP_TASK_SIZE) {
		return 1;
	}
	p = yk0a_set_priority_read_priority(link);
	if (p >= YK0A_SETUP_PRIORITY_SIZE) {
		return 1;
	}
	yk0a_set_priority_write_priority(link, yk0a_task_read_priority(yk0a_task_to_task_link(t)));
	if (yk0a_task_read_state(yk0a_task_to_task_link(t)) == YK0A_STATE_READY) {
		yk0a_schedule_delete(yk0a_task_to_task_link(t));
	}
	yk0a_task_write_priority(yk0a_task_to_task_link(t), p);
	if (yk0a_task_read_state(yk0a_task_to_task_link(t)) == YK0A_STATE_READY) {
		yk0a_schedule_append(yk0a_task_to_task_link(t));
		yk0a_schedule_update();
	}
	return 0;
}

static yk0a_error_t yk0a_procedure_interrupt_bit_mask(yk0a_task_link_t link) {
	yk0a_uint32_t i;
	yk0a_interrupt_t off;
	yk0a_interrupt_t on;
	off = yk0a_interrupt_bit_mask_read_mask_off(link);
	on = yk0a_interrupt_bit_mask_read_mask_on(link);
	i = yk0a_interrupt_read_mask();
	yk0a_interrupt_bit_mask_write_mask(link, i);
	yk0a_interrupt_write_mask((i & ~off) | on);
	return 0;
}

static yk0a_error_t yk0a_procedure_interrupt_bit_request(yk0a_task_link_t link) {
	yk0a_uint32_t i;
	yk0a_interrupt_t off;
	yk0a_interrupt_t on;
	off = yk0a_interrupt_bit_request_read_request_off(link);
	on = yk0a_interrupt_bit_request_read_request_on(link);
	i = yk0a_interrupt_read_request();
	yk0a_interrupt_bit_request_write_request(link, i);
	yk0a_interrupt_write_request((i & ~off) | on);
	return 0;
}

/* external function */

void yk0a_core_initialize(void) {
	yk0a_task_t t;
	yk0a_interrupt_initialize();
	yk0a_port_initialize();
	yk0a_schedule_initialize();
	for (t = 0; t < YK0A_SETUP_TASK_SIZE; ++t) {
		yk0a_task_write_priority(yk0a_task_to_task_link(t), 0);
		if (t == 0) {
			yk0a_task_write_state(yk0a_task_to_task_link(t), YK0A_STATE_READY);
			yk0a_schedule_append(yk0a_task_to_task_link(t));
		} else {
			yk0a_task_write_replyer(yk0a_task_to_task_link(t), yk0a_task_to_task_link(0));
			yk0a_task_write_state(yk0a_task_to_task_link(t), YK0A_STATE_REPLY);
		}
#if defined(YK0A_CUSTOM_ENABLE_PORT)
		yk0a_task_write_context(yk0a_task_to_task_link(t), YK0A_CUSTOM_CONTEXT_PORT, ~(yk0a_port_t)0);
#endif
	}
	yk0a_schedule_update();
}

void yk0a_core_exception_interrupt_procedure(void) {
	yk0a_uint32_t cause;
	yk0a_uint32_t i;
	yk0a_task_link_t l;
	cause = yk0a_mips_read_cause();
	switch (cause & YK0A_MIPS_CAUSE_EXCCODE_MASK) {
	case YK0A_MIPS_CAUSE_EXCCODE_INT:
		i = yk0a_interrupt_read_request();
		yk0a_interrupt_write_mask(yk0a_interrupt_read_mask() & ~i);
		yk0a_interrupt_bit_flag(YK0A_SETUP_INTERRUPT_PORT, YK0A_FLAG_INTERRUPT_MASK, i);
		break;
	case YK0A_MIPS_CAUSE_EXCCODE_SYS:
		l = yk0a_schedule_link;
		yk0a_task_write_context(l, YK0A_CONTEXT_EPC, yk0a_task_read_context(l, YK0A_CONTEXT_EPC) + 4); /* ignore YK0A_MIPS_CAUSE_BD */
#if defined(YK0A_CUSTOM_ENABLE_PORT)
		if (yk0a_task_read_context(l, YK0A_CUSTOM_CONTEXT_PORT) < YK0A_SETUP_PORT_SIZE) {
			yk0a_exception_forward(yk0a_task_read_context(l, YK0A_CUSTOM_CONTEXT_PORT), l);
		} else {
#endif
		switch (yk0a_task_read_context(l, YK0A_CONTEXT_ARGUMENT0)) {
		case YK0A_PROCEDURE_ACCEPT:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_accept(l));
			break;
		case YK0A_PROCEDURE_BIT_FLAG:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_bit_flag(l));
			break;
		case YK0A_PROCEDURE_CALL:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_call(l));
			break;
		case YK0A_PROCEDURE_CONNECT:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_connect(l));
			break;
		case YK0A_PROCEDURE_FORWARD:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_forward(l));
			break;
		case YK0A_PROCEDURE_GET_CONTEXT:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_get_context(l));
			break;
		case YK0A_PROCEDURE_GET_PRIORITY:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_get_priority(l));
			break;
		case YK0A_PROCEDURE_LISTEN:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_listen(l));
			break;
		case YK0A_PROCEDURE_REPLY:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_reply(l));
			break;
		case YK0A_PROCEDURE_SET_CONTEXT:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_set_context(l));
			break;
		case YK0A_PROCEDURE_SET_PRIORITY:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_set_priority(l));
			break;
		case YK0A_PROCEDURE_INTERRUPT_BIT_MASK:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_interrupt_bit_mask(l));
			break;
		case YK0A_PROCEDURE_INTERRUPT_BIT_REQUEST:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, yk0a_procedure_interrupt_bit_request(l));
			break;
		default:
			yk0a_task_write_context(l, YK0A_CONTEXT_ARGUMENT0, 1);
			break;
		}
#if defined(YK0A_CUSTOM_ENABLE_PORT)
		}
#endif
		break;
	default:
		yk0a_task_write_context(yk0a_schedule_link, YK0A_CONTEXT_CAUSE, cause);
#if defined(YK0A_CUSTOM_ENABLE_PORT)
		if (yk0a_task_read_context(yk0a_schedule_link, YK0A_CUSTOM_CONTEXT_PORT) < YK0A_SETUP_PORT_SIZE) {
			yk0a_exception_forward(yk0a_task_read_context(yk0a_schedule_link, YK0A_CUSTOM_CONTEXT_PORT), yk0a_schedule_link);
		} else {
#endif
		yk0a_exception_forward(YK0A_SETUP_EXCEPTION_PORT, yk0a_schedule_link);
#if defined(YK0A_CUSTOM_ENABLE_PORT)
		}
#endif
		break;
	}
}
